import os
import json
import logging
import datetime
import pytz

import tableauserverclient as TSC
from tableauhyperapi import TableName
import pantab


def get_tableau_server_credentials(**kwargs):
    """ infers and/or updates defaults for tableau server
    defaults comes from credentials.json filepath set as environment 'TABLEAU_SERVER_CREDENTIALS'
    providing below Keyword Arguments over rides defaults

    Keyword Arguments:
        serverUrl {str} - e.g. 'https://server.company.com'
        contentUrl {str} - siteName
        username {str} - userid
        password {str} 
        project {str} - tableau server folder/project ,defaults to 'Default' project
    

    Returns:
    {dict} - credential dictionary

    {'serverUrl': 'https://server.company.com' \
        , 'contentUrl': 'siteName' \
        , 'username' : 'userid' \
        , 'password': '***' \
        , 'project': 'Default'} 
    
    """
    default_credpath =  os.environ.get('TABLEAU_SERVER_CREDENTIALS')
    
    if default_credpath:
        with open(default_credpath, 'r') as file:
            credentials = json.load(file)
    else:
        logging.warning('default credentials path does not exit, no defaults set for tableau server authentication')
        credentials = {}
    
    if 'serverUrl' in kwargs:
        credentials['serverUrl'] = kwargs['serverUrl']
    
    if 'contentUrl' in kwargs:
        credentials['contentUrl'] = kwargs['contentUrl']
        
    if 'username' in kwargs:
        credentials['username'] = kwargs['username']
        
    if 'password' in kwargs:
        credentials['password'] = kwargs['password']
    
    if 'project' in kwargs:
        credentials['project'] = kwargs['project']
        
    return credentials

def upload_hyper_tableau_server(hyper_file, **kwargs):
    """ publishes local extract.hyper file to tableau server

    Arguments:
        hyper_file {str} -- filepath of extract.hyper or .tde to be published

    Keyword Arguments:
        project {str} -- tableau prep folder/project (Defaults: {Defaults})
        mode {str} -- CreateNew|Overwrite|Append (Defaults: Overwrite)
        embed {bool} -- if True, embeds credentails in data source (Defaults: True)
        name {str} -- name of extract file on server to be replaced or appended to
        ``serverUrl`` , ``contentUrl`` ,``username``, ``password``  can also be provide to override defaults for Tableau server authentication
    
    Returns:
        str -- datasource_name on the server
    """

    credentials = get_tableau_server_credentials(**kwargs)
    tableau_auth = __tableau_auth(credentials)
    server = TSC.Server(credentials.get('serverUrl'))
    project = credentials.get('project')


    logging.debug(f'supplied keyword args: {kwargs}')
    with server.auth.sign_in(tableau_auth):
        assert server.is_signed_in() == True

        all_project_items, pagination_item = server.projects.get()
        try:
            project_id = [
                item.id for item in all_project_items if item.name == project][0]
        except:
            project_id = [
                item.id for item in all_project_items if item.name == 'Default'][0]

        logging.debug(f'selected {project} with id {project_id}')

        if os.path.isfile(hyper_file):
            name = os.path.basename(hyper_file)
            hyper_filepath = hyper_file
        else:  # when only filename is passed which is in curren working directory folder
            name = hyper_file
            hyper_filepath = os.path.join(os.path.curdir, name)

        if 'name' in kwargs.keys() and kwargs.get('name') != None:
            name = kwargs.get('name')
        else:
            name = name.split('.')[0]

        if 'mode' in kwargs.keys():
            mode = kwargs.get('mode').title()
            if mode == 'Createnew':
                mode = 'CreateNew'

            if mode == 'Append':
                all_datasources, pagination_item = server.datasources.get()

                try:
                    datasource_id = [
                        datasource.id for datasource in all_datasources if datasource.name == name][0]
                    data_source_item = server.datasources.get_by_id(
                        datasource_id)
                    logging.debug(f"appending {hyper_filepath} to {name}")
                except:
                    mode = 'CreateNew'
                    data_source_item = TSC.DatasourceItem(
                        project_id=project_id, name=name)
                    logging.warning(
                        f'could not find {name} on server, publishing  {hyper_file} a new datasource')
            else:
                data_source_item = TSC.DatasourceItem(
                    project_id=project_id, name=name)
                logging.debug(
                    f"publishing {hyper_filepath} as {name} in {mode} mode")

        else:
            mode = 'Overwrite'
            data_source_item = TSC.DatasourceItem(
                project_id=project_id, name=name)
            logging.debug(f"publishing {name} in {mode} mode")

        if 'embed' in kwargs.keys():
            embedded_credential = TSC.ConnectionCredentials(
                username, password, embed=embed, oauth=False)
            server.datasources.publish(
                data_source_item, hyper_filepath, mode, connection_credentials=embedded_credential)
        else:
            server.datasources.publish(
                data_source_item, hyper_filepath, mode)
    logging.info(
        f'{hyper_filepath} was successfully published in {mode} mode to {project} project!')
    # TODO datasource_id or datasource_name not included in url at this time
    # url = __make_datasource_url(datasource_id, serverUrl=serverUrl, contentUrl=contentUrl)

    return name

def download_hyper_tableau_server(datasource_name, filepath=None, output_option='hyper', **credentials):
    """ downloads tdsx/hyper datasources from tableau server

    Arguments:
        datasource_name {str} -- datasource name without file extention

    Keyword Arguments:
        filepath {str} -- filepath for downloaded datasource (default: {None})
        output_option {str} -- to download ``tdsx``, set output_ouption = None (default: {'hyper'})
        kwargs for over-riding default credentials

    Returns:
        str -- filepath of downloaded data source in hyper ot tdsx format
    """
    credentials = get_tableau_server_credentials(**credentials)
    tableau_auth = __tableau_auth(credentials)
    datasource_id = __find_datasource_id_by_name(name = datasource_name)
    server = TSC.Server(credentials.get('serverUrl'))

    with server.auth.sign_in(tableau_auth):
            assert server.is_signed_in() == True
            data_source_item = server.datasources.get_by_id(datasource_id)
            filepath = server.datasources.download(datasource_id, filepath=filepath)

    if output_option == 'hyper':
        tdsx_path = filepath
        filepath = unzip_tdsx(filepath, hyper_filepath=None)
        os.remove(tdsx_path)
        
    return filepath

def df_to_hyper(df, filepath=None, table_name=None):
    """converts dataframe to hyper format on local drive, gcs, azure storage, tableau server, pass cred/clients?

    Arguments:
        df {pandas.DataFrame}
    Keyword Arguments:
        filepath {str} -- filename or filepath for hyper output (default: {extract_timestamp.hyper})
        table_name {tableauhyperapi.TableName} (default: {TableName('Extract', 'Extract')})

    Returns:
        str -- filepath of created hyper file
    """
    if not table_name:
        table_name = TableName('Extract', 'Extract')
        
    if not filepath:
        timezone = pytz.timezone("America/New_York")
        utc_now = pytz.utc.localize(datetime.datetime.utcnow())
        est_now_str = utc_now.astimezone(pytz.timezone(
            "America/New_York")).strftime("%Y%m%d_%H%M%S_EST")
        filepath = 'extract_' + est_now_str + '.hyper'
    else:
        filepath = filepath.split('.')[0] + '.hyper'
        
    pantab.frame_to_hyper(df, filepath, table=table_name)
    return filepath

def hyper_to_df(hyper_filepath, table_name=None):
    """ creates pandas.DataFrame form hyper

    Arguments:
        hyper_filepath {str} -- filepath of hyper file

    Keyword Arguments:
        table_name {tableauhyperapi.TableName} (default: {TableName('Extract', 'Extract')})

    Returns:
        pandas.DataFrame
    """
    if not table_name:
        table_name = TableName('Extract', 'Extract')
    return pantab.frame_from_hyper(hyper_filepath, table=table_name)

def __tableau_auth(credentials):
    """ instantiates tableau auth
    
    credentials {dict}
    """
    server_address = credentials.get('serverUrl')
    username = credentials.get('username')
    password = credentials.get('password')
    logging.info(f"singing into {server_address} as {username}")

    
    tableau_auth = TSC.TableauAuth(
        username, password, credentials.get('contentUrl'))
    return tableau_auth

def __find_datasource_id_by_name(name='extract.hyper', output_option=None, **kwargs):
    """ finds datasource_id on tableau server

    Keyword Arguments:
        name {str} -- datasource name (default: {'extract.hyper'})
        output_option {str} -- change output, if output_option = DICT returns dictionary of all datasource with name: id
                                if output_option = LIST returns list of all datasource attributes  (default: {None})

    Returns:
        str, dict, list
    """
    credentials = None
    if not credentials:
        credentials = get_tableau_server_credentials()
    tableau_auth = __tableau_auth(credentials)
    server = TSC.Server(credentials.get('serverUrl'))
    with server.auth.sign_in(tableau_auth):
        assert server.is_signed_in() == True
        
        all_datasources, pagination_item = server.datasources.get()
    
        if output_option == 'DICT':
            datasource_id = {datasource.name: datasource.id for datasource in all_datasources}
            return datasource_id
        elif output_option == 'LIST':
            datasource_id = [{'name' : datasource.name, 'id': datasource.id, 'project_name': datasource.project_name,'owner_id': datasource.owner_id, 'updated_at': datasource.updated_at , 'content_url': datasource.content_url} for datasource in all_datasources]
            return datasource_id

        else:
            datasource_id = [datasource.id for datasource in all_datasources if datasource.name == name]
            
        if len(datasource_id) >= 1:
            return datasource_id[0]
        else:
            logging.warning('{} datasource not found. Below is all datasources name and id \n {}'.format(name, {datasource.name: datasource.id for datasource in all_datasources}))
            

def unzip_tdsx(tdsx_filepath, hyper_filepath=None):
    """ extracts extract.hyper from tdsx

    Arguments:
        tdsx_filepath {str} -- filepath for tdsx

    Keyword Arguments:
        hyper_filepath {str} -- filepath for hyper (default: {tdsx_filepath})

    Returns:
        str -- filepath for hyper
    """
    import zipfile
    if not hyper_filepath:
        hyper_filepath = '.'.join(tdsx_filepath.split('.')[:-1])  + '.hyper'
        
    tdsx = zipfile.ZipFile(tdsx_filepath)
    for item in tdsx.namelist():
        if '.hyper' in item:
            f = tdsx.open(item)
            content = f.read()
            f = open(hyper_filepath, 'wb')
            f.write(content)
            f.close()
    return hyper_filepath


def __parse_datasource_url(datasource_url):
    #TODO datasource_id not part of url at this time ref: https://github.com/tableau/server-client-python/issues/268
    """ decomposes datasource url into components

    Arguments:
        datasource_url {str} -- url of extract.hyper file on tableau server e.g. 'https://server.company.com/#/site/contentUrl/project/163271/askData'

    Returns:
        tuple -- serverUrl, contentUrl, project, datasource_Id
    """
    url = datasource_url.split('/')
    serverUrl, contentUrl, project, datasource_Id = '/'.join(url[:3]), url[5], url[6] , url[7]
    return serverUrl, contentUrl, project, datasource_Id

def __make_datasource_url(datasource_id, **kwargs):
    #TODO datasource_id not part of url at this time ref https://github.com/tableau/server-client-python/issues/268
    """ creates datasource url from datasource_id
    Arguments:
        datasource_id {str}
    
    Keyword Arguments:
        serverUrl {str} - e.g. 'https://server.company.com'
        contentUrl {str} - siteName
        project {str} - tableau server folder/project ,defaults to 'Default' project

    Returns:
        url {str} -- 'https://server.company.com/#/site/contentUrl/project/163271/askData'
    """
    
    defaults_creds = get_tableau_server_credentials()
    
    if 'serverUrl' not in kwargs:
        serverUrl = defaults_creds.get('serverUrl')
    else:
        serverUrl = kwargs['serverUrl']
        
    if 'contentUrl' not in kwargs:
        contentUrl = defaults_creds.get('contentUrl')
    else:
        contentUrl = kwargs['contentUrl']
        
    if 'project' not in kwargs:
        project = defaults_creds.get('project')
    else:
        project = kwargs['project']
        
    url = serverUrl +'/#/site/' + contentUrl +'/' + project + '/' + datasource_id + '/askData'
    
    return url



__version__ = '1.0.0'



from .local_files import list_files_in_folder
from .gcp import df_to_bq , bq_to_df , download_blob_gcs , upload_blob_gcs, get_gcp_service_account
from .ms_azure import azure_to_df , df_to_azure_sql, azure_sql_con, azure_sql_engine, download_blob_azure_storage , upload_blob_azure_storage, get_azure_credentials
from .tableau_server import df_to_hyper, hyper_to_df, download_hyper_tableau_server, upload_hyper_tableau_server, get_tableau_server_credentials



import requests
from google.oauth2.service_account import Credentials
import google.oauth2.credentials
import google.auth.iam
from google.auth.transport.requests import Request




def gcf_authenticated_request(gcf_url, method='GET', credentials=None, IAM_SCOPE=None, **kwargs):
    """Makes authenticated request to google cloud functions
    code is copy-pasted and modified code from here:
    https://cloud.google.com/iap/docs/authentication-howto#authenticating_from_a_service_account

    Arguments:
        gcf_url {str} -- target google cloud function url e.g. https://us-central1-custom-ground-236517.cloudfunctions.net/hello_protected

    Keyword Arguments:
        method {str} -- request method e.g POST, GET (default: {'GET'})
        credentials {google.oauth2.service_account.Credentails } -- service account that has function.invoker permission on target url (default: {None})
        IAM_SCOPE {list} -- list of google iam scope (default: ['https://www.googleapis.com/auth/iam'])
        **kwargs -- arguemnts for requests.request method

    Returns:
        requests.response -- response object

    Example:
    gcf_url = 'https://us-central1-custom-ground-236517.cloudfunctions.net/hello_protected'
    resp = gcf_authenticated_request(gcf_url) # GET request

    # put request with **kwargs
    gcf_url = 'https://us-central1-custom-ground-236517.cloudfunctions.net/hello_protected'
    body ={"message":"this is a protected function, safe from crona virus "}
    headers = {'Content-Type':'application/json'}
    resp = gcf_authenticated_request(gcf_url,method='POST',headers=headers, data=json.dumps(body))
    """


    token_url = f'https://www.googleapis.com/oauth2/v1/tokeninfo?access_token=\\'
    metadata_server_token_url = 'http://metadata/computeMetadata/v1/instance/service-accounts/default/identity?audience='
    OAUTH_TOKEN_URI = 'https://www.googleapis.com/oauth2/v4/token'

    def get_google_open_id_connect_token(service_account_credentials):
        """Get an OpenID Connect token issued by Google for the service account.

        This function:

          1. Generates a JWT signed with the service account's private key
             containing a special "target_audience" claim.

          2. Sends it to the OAUTH_TOKEN_URI endpoint. Because the JWT in #1
             has a target_audience claim, that endpoint will respond with
             an OpenID Connect token for the service account -- in other words,
             a JWT signed by *Google*. The aud claim in this JWT will be
             set to the value from the target_audience claim in #1.

        For more information, see
        https://developers.google.com/identity/protocols/OAuth2ServiceAccount .
        The HTTP/REST example on that page describes the JWT structure and
        demonstrates how to call the token endpoint. (The example on that page
        shows how to get an OAuth2 access token; this code is using a
        modified version of it to get an OpenID Connect token.)
        """

        service_account_jwt = (
            service_account_credentials._make_authorization_grant_assertion())
        request = google.auth.transport.requests.Request()
        body = {
            'assertion': service_account_jwt,
            'grant_type': google.oauth2._client._JWT_GRANT_TYPE,
        }
        token_response = google.oauth2._client._token_endpoint_request(
            request, OAUTH_TOKEN_URI, body)
        return token_response['id_token']

    token_request_url = metadata_server_token_url + gcf_url
    token_request_headers = {'Metadata-Flavor': 'Google'}

    if not IAM_SCOPE:
        IAM_SCOPE = ['https://www.googleapis.com/auth/iam']

    if credentials:
        bootstrap_credentials = credentials
        bootstrap_credentials.has_scopes = IAM_SCOPE
    else:
        bootstrap_credentials, _ = google.auth.default(
            scopes=IAM_SCOPE)

    bootstrap_credentials.refresh(Request())
    signer = bootstrap_credentials.signer
    signer_email = bootstrap_credentials.service_account_email

    service_account_credentials = Credentials(
        signer, signer_email, token_uri=OAUTH_TOKEN_URI, additional_claims={
            'target_audience': gcf_url
        })
    headers = {}
    if 'headers' in kwargs.keys():
        for key in list(kwargs.keys()):
            if key == 'headers':
                headers = kwargs[key]
                del kwargs[key]

    google_open_id_connect_token = get_google_open_id_connect_token(
        service_account_credentials)

    headers['Authorization'] = f'Bearer {google_open_id_connect_token}'
    try:
        response = requests.request(method, gcf_url, headers=headers, **kwargs)
        return response
    except requests.exceptions.ReadTimeout:
        return print(f"triggered next gcf {gcf_url} but didn't wait for response")
    except:
        if response.status_code == 403:
            raise Exception(f'Service account {signer_email} does not have permission to '
                            'access the IAP-protected application.')
        elif response.status_code != 200:
            raise Exception(
                f'Bad response from application at {response.url}: {response.status_code} / {response.headers} / {response.text}')
        else:
            return response


import os
import json
import logging
import re


def list_files_in_folder(folderpath="./", file_extenion="."):
    """returns filepath in a local folder and its subfolders, filters filepaths by file extension

    Keyword Arguments:
        folderpath {str} -- relative or absolute path of directory (default: {"./"})
        file_extenion {str} -- e.g .py, .ipynb, .txt, .pdf etc (default: all filetype)

    Returns:
        list -- list of filepath matching the criteria
    """

    f = []

    if folderpath == "./":
        folderpath = os.path.curdir
        logging.debug("searching in current working dir and it's subfolders")
    else:
        logging.debug(f"searching in {folderpath} and it's subfolders")

    if file_extenion == ".":
        logging.debug("searching for all files")
    else:
        logging.debug(f"searching for files with {file_extenion} extension")
        file_extenion = file_extenion.lower()+"$"

    for (dirpath, dirnames, filenames) in os.walk(folderpath):
        for file in filenames:
            if re.search(r'{}'.format(file_extenion), file):
                f.append(os.path.join(dirpath, file))
    logging.debug(f"found {len(f)} files in")
    return f

def read_json_file(filepath):
    """ Reads json file

    Arguments:
        filepath {str} -- filepath

    Returns:
        dict
    """
    with open(filepath, mode='r') as file:
        content = json.loads(file.read())
    return content

def _update_json_file(filepath, newfilepath=None, **kwargs):
    """ Updates or creates newfile json file from existing one

    Arguments:
        filepath {str} -- json filepath to be updated

    Keyword Arguments:
        newfilepath {str} -- new filepath so as not modify original file (default: {overwrites original file})
        kwargs:
        key = value pairs or dict to be inserted or updated in the original file

    Returns:
        newfilepath {str} -- modified filepath
    """
    if not newfilepath:
        newfilepath = filepath
    with open(filepath,mode = 'r') as file:
        dict_content = json.loads(file.read())
 
    for key in kwargs:
        dict_content[key] = kwargs[key]

    with open(newfilepath,mode = 'w') as file:
        file.write(json.dumps(dict_content))
    
    return newfilepath

def _delete_temp_files():
    if os.path.isfile(temp_azure_cred):
        os.remove(temp_azure_cred)

    if os.path.isfile(temp_tableau_cred):
        os.remove(temp_tableau_cred)

    if os.path.isfile(temp_gcp_cred):
        os.remove(temp_gcp_cred)

def _insert_into_namespace(key, value, name_space=globals()):
    """insert key, values to global namesapce

    Arguments:
        key {str} 
        value {any} 

    Keyword Arguments:
        name_space -- (default: {globals()})
    """
    name_space[key] = value


import os
import io
import logging
import json
import pandas as pd
import datetime
import pytz
from google.cloud import bigquery
from google.cloud import storage

def download_blob_gcs(gcs_uri, filepath=None, storage_client=None, output_option='FILE', delete_blob=False):
    """ Downloads blob from Google Cloud Storage to local drive or buffer

    Arguments:
        gcs_uri {str} -- google cloud storage blob uri e.g. gs://bucketname/foldername/blobname.csv

    Arguments:
        filepath {str} -- folder_path/new_filename (default: current_working_directory/blob_name)
        storage_client {google.storage.Client} -- storage client instantiated with non-default credentials (default: {None})
        output_option {str} -- determines function output type. (default: {'FILE'})
        delete_blob {bool} -- flag for deleting the blob after download is successful (default: {False})

    Returns:
        output_option : return type
        {STRING : unicode string}
        {IO : io.BytesIO}
        {URL: downloadable link}
        {CREDENTIALS: google.oauth2.service_account.Credentials}
        {JSON: DICT}
        {FILE: filepath} 
    """
    from google.oauth2.service_account import Credentials
    
    if not storage_client:
        logging.debug(
        "instantiating storage client from defualt environment variable")
        storage_client = storage.Client()

    bucket = gcs_uri[5:].split('/')[0]
    blob = '/'.join(gcs_uri[5:].split('/')[1:])
    blob = storage_client.get_bucket(bucket).get_blob(blob)
    
    logging.debug(
    f"storage client {storage_client.get_service_account_email()} is requesting access to blob: {blob} at {gcs_uri} ")

    if output_option == 'IO':
        output = io.BytesIO()
        blob.download_to_file(output)
        output.seek(0)

    elif output_option == 'STRING':
        output = blob.download_as_string().decode()
    
    elif output_option == 'JSON':
        output = json.loads(blob.download_as_string().decode())

    
    elif output_option == 'CREDENTIALS':
        output = Credentials.from_service_account_info(json.loads(blob.download_as_string().decode()))
    
    elif output_option == 'URL':
        expiration_duration = datetime.timedelta(weeks=1)
        output = blob.generate_signed_url(expiration_duration)
        

    else:
        if not filepath:
            filepath = gcs_uri.split("/")[-1]
        with open(filepath, mode="wb") as file_obj:
            storage_client.download_blob_to_file(gcs_uri, file_obj)
        output = filepath

    if delete_blob:
        blob.delete()

    return output

def upload_blob_gcs(content, bucket_id=None, blobname=None,storage_client=None):
    """ Uploads content to Google Cloud Storage

    Arguments:
        content {str, filepath, io.BytesIO} -- string, filepath or BytesIO to be uploaded

    Keyword Arguments:
        bucket_id {str} -- google cloud storage bucket id (default: default bucket from env variable "STORAGE_BUCKET")
        blobname {str} -- blob name, for string and BytesIO name should be provided (default: {unname_blob_uploaded_at_timestamp})
        storage_client {google.storage.Client} -- storage client instantiated with non-default credentials (default: {None})

    Return:
        {str} -- returns google cloud storage uri of uploaded blob
    """

    if not storage_client:
        logging.debug(
        "instantiating storage client from defualt environment variable")
        storage_client = storage.Client()
        
    if not bucket_id:
        bucket_id = os.environ.get("STORAGE_BUCKET")
    
    if not blobname:
        if os.path.isfile(content):
            blobname = os.path.basename(content)
            storage_client.get_bucket(bucket_id).blob(blobname).upload_from_filename(content)
            return f'gs://{bucket_id}/{blobname}'
        else:
            blobname = 'unname_blob_uploaded_at' + datetime.datetime.now().isoformat().replace('-','_').replace(':','_')[:19]

    if isinstance(content, io.BytesIO):
        storage_client.get_bucket(bucket_id).blob(blobname).upload_from_file(content)
    else:
        storage_client.get_bucket(bucket_id).blob(blobname).upload_from_string(content)

    return f'gs://{bucket_id}/{blobname}'

def bq_to_df(sql, client=None, **job_config):
    """ Returns bigquery query result as pandas dataframe

    Arguments:
        sql {str} -- bigquery SELECT statement in standard SQL or stored procedure containing SELECT statement

    Keyword Arguments:
        client {bigquery.Client} -- defaults to client instantiated with default credentials
        job_config {dict} -- keyword arguemnt for bigquery.QueryJob.Config()

    Returns:
        pandas.DataFrame -- query result as df
    """
    if not client:
        logging.debug(
        "instantiating bigquery client from defualt environment variable")
        client = bigquery.Client()

    def get_date_columns(job):
        """ returns list of DATE columns from bigquery.job object
        """
        return [field.name for field in job.result().schema if field.field_type == 'DATE']

    def transform_date(date_colums):
        """ castes datetime.date to pd.datetime as pantab doesn't handle datetime.date objects"""
        for col in date_colums:
            df[col] = pd.to_datetime(df[col])

    job_config = bigquery.QueryJobConfig(**job_config)

    dry_run = bigquery.QueryJobConfig(
        dry_run=True, use_query_cache=False)
    job = client.query(sql, job_config=dry_run)

    if job.statement_type == 'SELECT':
        job_id = create_bq_job_id(
            'hyper publiser app SELECT Statment request')

        job = client.query(sql, job_id=job_id, job_config=job_config)
        # job.result()
        df = job.to_dataframe()
        # job_info = get_job_info(job,client=client)
        date_colums = get_date_columns(job)

        if date_colums:
            transform_date(date_colums)

    elif job.statement_type == 'SCRIPT':
        job_id = create_bq_job_id(
            'hyper publiser app script request')
        job = client.query(sql, job_id=job_id, job_config=job_config)
        job.result()
        job_id = get_job_info(job,client=client, output_option='LIST')

        if len(job_id) == 1:
            df = client.get_job(job_id[0]).to_dataframe()
            # job_info = get_job_info(job,client=client)
            date_colums = get_date_columns(job)

            if date_colums:
                transform_date(date_colums)

        elif len(job_id) > 1:
            df = client.get_job(job_id[-1]).to_dataframe()
            # job_info = get_job_info(job,client=client)
            logging.waring(
                " multi select stored procedure returns data for the last SELECT statement only")
            date_colums = get_date_columns(job)

            if date_colums:
                transform_date(date_colums)

        else:
            logging.error(f"{sql} script did not return any data")
            df = None
    else:
        logging.error(
            f"{job.statement_type} are not suppoted. Please provide a SELECT statement or a stored procedure containing SELECT statement")
        return
        # job_info = get_job_info(job,client=client)
        # logging.info(f"job execution detail: {job_info}")
    logging.info(f"{job.statement_type} statement returned {len(df)} rows")

    return df

def df_to_bq(df, table_id, client=None, write_mode='WRITE_APPEND', schema=None, autodetect=False, job_id=None,**job_config):
    """ loads dataframe to bigquery with custom datatype and schema

    Arguments:
        df {pd.DataFrame} -- pandas dataframe to be updoad
        table_id {str} -- fully qualified bq table id as write destination e.g. project_id.dataset.new_tablename
        client {bigquery.Client} -- defaults to client instantiated with default credentials
        write_mode {str} -- optional argument. Valid values are 'WRITE_APPEND' , 'WRITE_TRUNCATE' OR 'WRITE_EMPTY'. defaults to WRITE_APPEND mode
        schema {bigquery table schema} -- optional, schema of bigquery table, use get_table_schema(table_id) to enforce existing table schema
        autodetect {bool} -- optional, if True, infers datatype, deaults to False
        job_id {str} -- optional argument, use function create_bq_job_id(description=None) to create custom job id

    Returns:
        Biqquery job object
        """
    if not client:
        logging.debug(
        "instantiating bigquery client from defualt environment variable")
        client = bigquery.Client()

    job_config = bigquery.LoadJobConfig(**job_config)
    job_config.autodetect = autodetect
    job_config.write_disposition = write_mode
    job_config.create_disposition = 'CREATE_IF_NEEDED'
    if schema:
        job_config.schema = schema

    load_job = client.load_table_from_dataframe(
        df, table_id, job_id=job_id, job_config=job_config)
    return load_job


def create_bq_job_id(description=None):
    """creates custom job id for bigquery jobs (load job, export, query, DML statements etc) with yyymmdd_hhmmss_EST_description pattern

    Arguments:
        description {str} -- description of job, use searchable terms to pull the job from logs (default: service_account_email prefix)

    Returns:
        job_id {str} -- custom job Id with yyymmdd_hhmmss_EST_description pattern
    """
    timezone = pytz.timezone("America/New_York")
    utc_now = pytz.utc.localize(datetime.datetime.utcnow())
    est_now_str = utc_now.astimezone(pytz.timezone(
        "America/New_York")).strftime("%Y%m%d_%H%M%S_EST")

    if description:
        description = description.replace(' ', '_')
        job_id = est_now_str + f"_{description}"
    else:
        agent = client.get_service_account_email()
        job_id = est_now_str + f"_{agent.split('@')[0]}"
    return job_id


def get_table_schema(table_id, client=None, output_option='OBJECT'):
    """Returns table schema of Biqquery Table

    Arguments:
        table_id {str} -- fully qualified table id e.g. project_id.dataset.new_tablename

    Keyword Arguments:
        client {bigquery.Client} -- defaults to client instantiated with default credentials
        output_option {str} -- determines return data type {'OBJECT'|'LIST'|'DICT'} where 'object' is bigquery.table.schema object
        , 'list' returns list of columns names, 'dict' returns list of dictionary for each field : field_type. (default: bigquery schema object)

    Returns:
        biqquery schema object|list of str|list of dict containing fields definition
    """
    if not client:
        logging.debug(
        "instantiating bigquery client from defualt environment variable")
        client = bigquery.Client()

    if output_option == 'LIST':
        return [field.name for field in client.get_table(table_id).schema]
    elif output_option == 'DICT':
        return [{field.name: field.field_type for field in client.get_table(table_id).schema}]
    else:
        return client.get_table(table_id).schema


def get_job_info(job, client=None, output_option=None):
    """ For query jobs: returns job_id, statement_type, number_of_row_affected, destination_table_path and children job info
    For laod jobs: return job_id, job_type, destination_table, write_mode, number of output rows and errors encountered

    Arguments:
        job {bigquery job object} -- query or load job

    Keyword Arguments:
        client {bigquery.Client} -- defaults to client instantiated with default credentials
        output_option {str} -- optional output for SCRIPT statement. 'LIST' retruns filtered list of children job_id of SELECT type where result total_row > 1. 'DICT' returns details of children jobs of SELECT type (default: {None})

    Returns:
        dict | list -- job info is dict, while children job info is list

    Example:
    get_job_info(client.query(sql_statement)) # returns job_info

    # for SCRIPT, returns detail of children jobs of SELECT statements
    get_job_info(script_job, output_option='DICT')
    """
    if not client:
        logging.debug(
        "instantiating bigquery client from defualt environment variable")
        client = bigquery.Client()

    def extract_table_id(job_destination_path):
        return f"{job_destination_path.split('/')[2]}.{job_destination_path.split('/')[4]}.{job_destination_path.split('/')[6]}"

    if job.job_type == 'query':
        if job.statement_type == 'SCRIPT':
            child_jobs_iterable = client.list_jobs(parent_job=job)
            children = [{'job_id': job.job_id, 'statement_type': job.statement_type, 'destination': extract_table_id(job.destination.path),
                         'num_dml_affected_rows': job.num_dml_affected_rows} if job.statement_type != 'SELECT' else {'job_id': job.job_id, 'statement_type': job.statement_type,
                                                                                                                     'total_rows': job.result().total_rows, 'destination': extract_table_id(job.destination.path), 'schema': get_table_schema(extract_table_id(job.destination.path), client=client, output_option='DICT')} for job in child_jobs_iterable]
            data = {'job_id': job.job_id, 'statement_type': job.statement_type,
                    'num-child_jobs': job.num_child_jobs, 'child_job': children}

            if output_option == 'LIST':
                children_job_id_SELECT = [job.get('job_id') for job in data.get(
                    'child_job') if job.get('statement_type') == 'SELECT' and job.get('total_rows') > 1]
                data = children_job_id_SELECT

            elif output_option == 'DICT':
                children_job_id_SELECT = [job.get('job_id') for job in data.get(
                    'child_job') if job.get('statement_type') == 'SELECT']
                data = [{'job_id': child, 'statement_type': 'SELECT', 'total_rows': client.get_job(child).result(
                ).total_rows, 'destination': extract_table_id(client.get_job(child).destination.path)} for child in children_job_id_SELECT]

        elif job.statement_type == 'SELECT':
            data = {'job_id': job.job_id, 'statement_type': job.statement_type,
                    'total_rows': job.result().total_rows, 'destination': extract_table_id(job.destination.path), 'schema': get_table_schema(extract_table_id(job.destination.path), client=client,output_option='DICT')}
        else:
            data = {'job_id': job.job_id, 'statement_type': job.statement_type,
                    'destination': extract_table_id(job.destination.path), 'num_dml_affected_rows': job.num_dml_affected_rows}
    else:
        data = {'job_id': job.job_id, 'job_type': job.job_type,
                'destination': extract_table_id(job.destination.path), 'write_mode': job.write_disposition, 'output_rows': job.output_rows, 'errors': job.errors}
    return data


def get_gcp_service_account(credentials):
    """ returns google.oauth2.service_account.Credentials from local filepath, gcs_uri or dict

    Arguemnts:
    credentials {str, bytes, dict}
    """
    from google.oauth2.service_account import Credentials
    if isinstance(credentials,dict):
        credentials = Credentials.from_service_account_info(credentials)
        logging.debug(f"project_id: {credentials.project_id} service_account_email: {credentials.service_account_email}")
    else:
        if os.path.isfile(credentials):
            credentials = Credentials.from_service_account_file(credentials)
            logging.debug(f"project_id: {credentials.project_id} service_account_email: {credentials.service_account_email}")
            
        elif isinstance(credentials,bytes):
            credentials = Credentials.from_service_account_info(json.loads(credentials))
            logging.debug(f"project_id: {credentials.project_id} service_account_email: {credentials.service_account_email}")

        elif 'gs://' in credentials:
            credentials = download_blob_gcs(credentials,output_option='CREDENTIALS')
            logging.debug(f"project_id: {credentials.project_id} service_account_email: {credentials.service_account_email}")
        else:
            credentials = Credentials.from_service_account_info(json.loads(credentials))
            logging.debug(f"project_id: {credentials.project_id} service_account_email: {credentials.service_account_email}")

    return credentials



import os
import json
import sys
import logging
import pandas as pd
from pyodbc import connect
from azure.storage.blob.blockblobservice import BlockBlobService

def get_azure_credentials(**kwargs):
    """ infers and/or updates defaults for Azure SQL and Azure storage account
    defaults comes from credentials.json filepath set as environment 'AZURE_CREDENTIALS'

    Keyword Arguments:
        account_name {str} 'https://{account_name}.blob.core.windows.net/{container_name}/{blob_name}'
        container_name {str} - storage container name
        account_key {str} - account_key for storage authentication

        server {str} -- azure sql server url e.g "myazuresql.database.windows.net"
        username {str} -- user id (default: admin )
        password {str} -- password (default: admin_password)
        database {str} -- database name (default: dev database)


    Returns:
    {dict} - credential dictionary

    """
    default_credpath =  os.environ.get('AZURE_CREDENTIALS')
    
    if default_credpath:
        with open(default_credpath, 'r') as file:
            credentials = json.load(file)
    else:
        logging.warning('default credentials path does not exit, no defaults set for tableau server authentication')
        credentials = {}
    
    if 'account_name' in kwargs:
        credentials['account_name'] = kwargs['account_name']
    
    if 'container_name' in kwargs:
        credentials['container_name'] = kwargs['container_name']

    if 'account_key' in kwargs:
        credentials['account_key'] = kwargs['account_key']


    if 'server' in kwargs:
        credentials['server'] = kwargs['server']
        
    if 'username' in kwargs:
        credentials['username'] = kwargs['username']
        
    if 'password' in kwargs:
        credentials['password'] = kwargs['password']
    
    if 'database' in kwargs:
        credentials['database'] = kwargs['database']
        
    return credentials

def download_blob_azure_storage(blob_url, account_key=None, filepath=None, output_option='FILE', delete_blob=False):
    """ downloads azure blob to FILE, io.BytesIO, STRING, JSON, google CREDENTIAL objects type


    Arguments:
        blob_url {str} -- azure storage blob url e.g https://account_name.blob.core.windows.net/container_name/blob_name
        account_key {str} -- authentication method for azure storage

    Keyword Arguments:
        filepath {str} -- folder_path/new_filename (default: current_working_directory/blob_name)
        output_option {str} -- determines function output type. (default: {'FILE'})
        delete_blob {bool} -- flag for deleting the blob after download is successful (default: {False})

    Returns:
        Returns:
        output_option : return type
        {STRING : unicode string}
        {IO : io.BytesIO}
        {URL: downloadable link}
        {CREDENTIALS: google.oauth2.service_account.Credentials}
        {JSON: DICT}
        {FILE: filepath} 
    """
    from io import BytesIO
    from google.oauth2.service_account import Credentials


    account_name, container_name, blob_name = __parse_storage_url(blob_url)
    if not account_key:
        account_key = get_azure_credentials().get('account_key')
        
    service = BlockBlobService(
        account_name=account_name, account_key=account_key)
    blob = service.get_blob_to_bytes(container_name, blob_name)

    if output_option == 'IO':
        output = BytesIO(blob.content)
        output.seek(0)

    elif output_option == 'STRING':
        output = blob.content.decode()
    
    elif output_option == 'JSON':
        output = json.loads(blob.content.decode())
    
    elif output_option == 'CREDENTIALS':
        output = Credentials.from_service_account_info(json.loads(blob.content.decode()))
    
    elif output_option == 'URL':
        from datetime import datetime, timedelta
        from azure.storage.blob import BlobPermissions
        from urllib.parse import quote

        sas_token = service.generate_blob_shared_access_signature(
            container_name=container_name
            ,blob_name=blob_name
            ,permission=BlobPermissions(read=True),
            expiry=datetime.utcnow() + timedelta(weeks=1)
        )
        output = service.make_blob_url(container_name, quote(blob_name),sas_token=sas_token)

    else:
        if not filepath:
            filepath = blob_name
        with open(filepath, mode='wb') as file:
            file.write(blob.content)
        output = filepath
    
    if delete_blob:
        service.delete_blob(container_name, blob_name)

    return output

def upload_blob_azure_storage(content, blob_name=None, **credentials):
    """Uploads content to Azure Storage

    Arguments:
        content {str, filepath, bytes, io.BytesIO} -- string, bytes, filepath or BytesIO to be uploaded


    Keyword Arguments:
        blob_name {str} -- blob name, for string and BytesIO name should be provided (default: {unname_blob_uploaded_at_timestamp})

    Returns:
        {str} -- returns Azure storage url of uploaded blob
    """
    import io
    from datetime import datetime
    credentials = get_azure_credentials(**credentials)
    
    account_name = credentials.get('account_name')
    container_name = credentials.get('container_name')
    account_key = credentials.get('account_key')
    
    service = BlockBlobService(
        account_name=account_name, account_key=account_key)
    
    if not blob_name:
        if os.path.isfile(content):
            blob_name= os.path.basename(content)
            service.create_blob_from_path(container_name,blob_name,content)
            return __make_storage_url(account_name,container_name,blob_name)
        else:
            blob_name = 'unname_blob_uploaded_at' + datetime.now().isoformat().replace('-','_').replace(':','_')[:19]

    if isinstance(content,io.BytesIO):
        service.create_blob_from_stream(container_name,blob_name,content)   
    elif isinstance(content,bytes):
        service.create_blob_from_bytes(container_name,blob_name,content)
    else:
        service.create_blob_from_text(container_name,blob_name,content)
    return __make_storage_url(account_name, container_name, blob_name)
    
def azure_sql_con(**credentials):
    """creates connection to azure sql. use the connection with pandas.read_sql(sql_query,connection)

    Keyword Arguments:
        to override defaults set in environment variable 'AZURE_CREDENTIALS' below keyword arg should be provided
        server {str} -- azure sql server url e.g "myazuresql.database.windows.net" (default:)
        username {str} -- user id (default: admin )
        password {str} -- password (default: admin_password)
        database {str} -- database name (default: dev database)

    Returns:
        pyodbc connection to microsoft azure sql
    """
    credentials = get_azure_credentials(**credentials)
    
    SERVER = credentials.get('server')
    database = credentials.get('database')
    username = credentials.get('username')
    password = credentials.get('password')

    try:
        connection = connect('DRIVER={};SERVER=tcp:{};database={};UID={};PWD={}'.format(
            'ODBC Driver 17 for SQL Server', SERVER, database, username, password))
        logging.info('connected to {} as {}'.format(database, username))
        return connection

    except Exception as e:
        logging.error(e)
        sys.exit('connection to azure sql failed')


def azure_to_df(sql_query, connection=None, **kwargs):
    """ Return result of sql query as pandas dataframe

    Arguments:
        sql_query {str} -- valid SELECT sql statement
        connection {pyodbc connection} -- uses azure_sql_con() function to create connection object
        kwargs {dict} -- key words arguments for pandas.read_sql function

    Returns:
        pandas.DataFrame
    """
    if not connection:
        try:
            connection = azure_sql_con()
        except Exception as e:
            logging.error(
                e)
            logging.error(
                "not connected to azure, please use azure_sql_con() function to create connection object and pass it as a parameter to the function")
    return pd.read_sql(sql_query, connection, **kwargs)

def azure_sql_engine(**credentials):
    """creates sqlalchemy engine to write to azure sql. use the engine with df_to_azure_sql(df,table_name,engine)

    Keyword Arguments:
        to override defaults set in environment variable 'AZURE_CREDENTIALS' below keyword arg should be provided
        server {str} -- azure sql server url e.g "myazuresql.database.windows.net" (default: azure_sql_server)
        username {str} -- user id (default: admin )
        password {str} -- password (default: admin_password)
        database {str} -- database name (default: dev database)

    Returns:
        sqlalchemy engine to microsoft azure sql
    """
    import sqlalchemy
    from urllib.parse import quote_plus
    credentials = get_azure_credentials(**credentials)
    
    server = credentials.get('server')
    database = credentials.get('database')
    username = credentials.get('username')
    password = credentials.get('password')

    connection_string = quote_plus('DRIVER={ODBC Driver 17 for SQL Server}'+f';SERVER=tcp:{server};DATABASE={database};UID={username};PWD={password}')
    engine = sqlalchemy.create_engine(f"mssql+pyodbc:///?odbc_connect={connection_string}")
    logging.info('connected to {} as {}'.format(database, username))
    return engine

def df_to_azure_sql(df, table_name, engine=None, if_exists='fail', index=False, **kwargs):
    """ writes dataframe to Azure sql table
    table_name = 
    if_exists=
    

    Arguments:
        df {pandas.DataFrame} -- data to be written to table
        table_name {str} -- table name e.g 'dbo.fact_sample_superstore'

    Keyword Arguments:
        engine {sqlalchemy.engine} --  (default: {None})
        if_exists {str} -- write mode {'fail', 'replace', 'append'} (default: {'fail'})
        index {bool} -- if true, writes dataframe index as well (default: {False})
        keyword arguments for df.to_sql()

    Returns:
        str -- destination table_name
    """
    
    if not engine:
        logging.debug('instantiating sql achemy engine with default credentials')
        engine = azure_sql_engine()
        
    df.to_sql(table_name,engine,index=index,if_exists=if_exists, **kwargs)
    return table_name

def __parse_storage_url(blob_url):
    account_name, container_name, blob_name = blob_url[8:].split('/')
    account_name = account_name.split('.')[0]
    return account_name, container_name, blob_name

def __make_storage_url(account_name,container_name,blob_name):
    return f'https://{account_name}.blob.core.windows.net/{container_name}/{blob_name}'



"""save the script on google cloud storage bucket and provide gcs_uri to pivotal app""

import requests
import pandas as pd
from io import BytesIO


def create_df():
    """requires data from http and returns dataframe to be converted to hyper

    Returns:
        pandas dataframe
    """

    url = 'https://archive.ics.uci.edu/ml/machine-learning-databases/abalone/abalone.data'
    response = requests.request('GET', url, verify=False)
    print(
        f'response code for request to {response.url} is {response.status_code}')
    in_mem_file = BytesIO(response.content)
    in_mem_file.seek(0)
    col_names = ['Sex', 'Length', 'Diameter', 'Height', 'Whole_weight',
                 'Shucked_weight', 'Viscera_weight', 'Shell_weight', 'Rings']
    return pd.read_csv(in_mem_file, sep=',', header=None, names=col_names)



from setuptools import setup, find_packages
import re
import os
import sys
import codecs

with open('requirements.txt') as f:
    required = f.read().splitlines()

try:
    # https://stackoverflow.com/questions/30700166/python-open-file-error
    with codecs.open("README.md", 'r', errors='ignore') as file:
        readme_contents = file.read()

except Exception as error:
    readme_contents = ""
    sys.stderr.write("Warning: Could not open README.md due %s\n" % error)


setup(name='hyperdb',
      version='0.1.1',
      description='Hyperdb provides wrapper functions for working with Tableau hyper datasources and moving data between Tableau Server, Google Cloud Platform and Microsoft Azure through a common interface',
      long_description=readme_contents,
      long_description_content_type="text/markdown",
      url='http://github.com/mhadi813/hyperdb',
      author='Muhammad Hadi',
      author_email='mhadi813@gmail.com',
      license='BSD',
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "Topic :: Office/Business",
          "License :: OSI Approved :: BSD License",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
      ],
      packages=find_packages(),
      install_requires=required,
      keywords="tableau hyper pandas google bigquery cloud storage azure sql database",
      zip_safe=False)
